# -*- coding: utf-8 -*-
"""Classes for Windows Registry access."""

from dfwinreg import definitions
from dfwinreg import key_paths
from dfwinreg import regf
from dfwinreg import virtual


class WinRegistryFileMapping(object):
  """Windows Registry file mapping.

  Attributes:
    key_path_prefix (str): Windows Registry key path prefix.
    unique_key_paths (list[str]): key paths unique to the Windows Registry file.
    windows_path (str): Windows path to the Windows Registry file, such as:
        C:\\Windows\\System32\\config\\SYSTEM
  """

  def __init__(self, key_path_prefix, windows_path, unique_key_paths):
    """Initializes the Windows Registry file mapping.

    Args:
      key_path_prefix (str): Windows Registry key path prefix.
      windows_path (str): Windows path to the Windows Registry file, such as:
          C:\\Windows\\System32\\config\\SYSTEM
      unique_key_paths (list[str]): key paths unique to the Windows Registry
          file.
    """
    super(WinRegistryFileMapping, self).__init__()
    self.key_path_prefix = key_path_prefix
    self.unique_key_paths = unique_key_paths
    self.windows_path = windows_path


class WinRegistry(object):
  """Windows Registry."""

  _REGISTRY_FILE_MAPPINGS = [
      # TODO: Windows 3.1
      # Windows 9x/Me
      WinRegistryFileMapping(
          'HKEY_LOCAL_MACHINE',
          '%SystemRoot%\\SYSTEM.DAT',
          ['\\Config', '\\Enum', '\\Hardware', '\\Network', '\\Software',
           '\\System']),
      WinRegistryFileMapping(
          'HKEY_USERS',
          '%SystemRoot%\\USER.DAT',
          ['\\.DEFAULT\\AppEvents', '\\.DEFAULT\\Control Panel',
           '\\.DEFAULT\\Keyboard Layout', '\\.DEFAULT\\Network',
           '\\.DEFAULT\\Software']),
      # Windows NT
      WinRegistryFileMapping(
          'HKEY_CURRENT_USER',
          '%UserProfile%\\NTUSER.DAT',
          ['\\AppEvents', '\\Console', '\\Control Panel', '\\Environment',
           '\\Keyboard Layout', '\\Software']),
      # Windows Vista and later UsrClass.dat mapping
      WinRegistryFileMapping(
          'HKEY_CURRENT_USER\\Software\\Classes',
          '%UserProfile%\\AppData\\Local\\Microsoft\\Windows\\UsrClass.dat',
          ['\\Local Settings\\Software\\Microsoft\\Windows\\CurrentVersion']),
      # Windows 2000, XP and 2003 UsrClass.dat mapping
      WinRegistryFileMapping(
          'HKEY_CURRENT_USER\\Software\\Classes',
          ('%UserProfile%\\Local Settings\\Application Data\\Microsoft\\'
           'Windows\\UsrClass.dat'),
          ['\\Software\\Microsoft\\Windows\\CurrentVersion']),
      WinRegistryFileMapping(
          'HKEY_LOCAL_MACHINE\\SAM',
          '%SystemRoot%\\System32\\config\\SAM',
          ['\\SAM\\Domains\\Account\\Users']),
      WinRegistryFileMapping(
          'HKEY_LOCAL_MACHINE\\Security',
          '%SystemRoot%\\System32\\config\\SECURITY',
          ['\\Policy\\PolAdtEv']),
      WinRegistryFileMapping(
          'HKEY_LOCAL_MACHINE\\Software',
          '%SystemRoot%\\System32\\config\\SOFTWARE',
          ['\\Microsoft\\Windows\\CurrentVersion\\App Paths']),
      WinRegistryFileMapping(
          'HKEY_LOCAL_MACHINE\\System',
          '%SystemRoot%\\System32\\config\\SYSTEM',
          ['\\MountedDevices', '\\Select', '\\Setup'])]

  _ROOT_KEY_ALIASES = {
      'HKCC': 'HKEY_CURRENT_CONFIG',
      'HKCR': 'HKEY_CLASSES_ROOT',
      'HKCU': 'HKEY_CURRENT_USER',
      'HKLM': 'HKEY_LOCAL_MACHINE',
      'HKU': 'HKEY_USERS'}

  _ROOT_SUB_KEYS = frozenset([
      'HKEY_CLASSES_ROOT',
      'HKEY_CURRENT_CONFIG',
      'HKEY_CURRENT_USER',
      'HKEY_DYN_DATA',
      'HKEY_LOCAL_MACHINE',
      'HKEY_PERFORMANCE_DATA',
      'HKEY_USERS'])

  _LOCAL_MACHINE_SUB_KEYS = frozenset([
      'SAM',
      'Security',
      'Software',
      'System'])

  _SELECT_KEY_PATH = 'HKEY_LOCAL_MACHINE\\System\\Select'

  _USER_PROFILE_LIST_KEY_PATH = (
      'HKEY_LOCAL_MACHINE\\Software\\Microsoft\\Windows NT\\CurrentVersion\\'
      'ProfileList')

  _USER_SOFTWARE_CLASSES_KEY_PATH = 'HKEY_CURRENT_USER\\SOFTWARE\\CLASSES'

  # TODO: add support for HKEY_CLASSES_ROOT
  # TODO: add support for HKEY_CURRENT_CONFIG
  # TODO: add support for HKEY_CURRENT_USER
  # TODO: add support for HKEY_DYN_DATA

  _VIRTUAL_KEYS = [
      ('HKEY_USERS', '_GetUsersVirtualKey')]

  def __init__(self, ascii_codepage='cp1252', registry_file_reader=None):
    """Initializes the Windows Registry.

    Args:
      ascii_codepage (Optional[str]): ASCII string codepage.
      registry_file_reader (Optional[WinRegistryFileReader]): Windows Registry
          file reader.
    """
    super(WinRegistry, self).__init__()
    self._ascii_codepage = ascii_codepage
    self._registry_file_reader = registry_file_reader
    self._registry_files = {}
    self._root_key = None
    self._user_registry_files = {}

  def __del__(self):
    """Cleans up the Windows Registry object."""
    for key_path_prefix_upper, registry_file in self._registry_files.items():
      self._registry_files[key_path_prefix_upper] = None
      if registry_file:
        registry_file.Close()

    for profile_path_upper, registry_file in self._user_registry_files.items():
      self._user_registry_files[profile_path_upper] = None
      if registry_file:
        registry_file.Close()

  def _GetCachedFileByPath(self, key_path_upper):
    """Retrieves a cached Windows Registry file for a key path.

    Args:
      key_path_upper (str): Windows Registry key path, in upper case with
          a resolved root key alias.

    Returns:
      tuple: consist:

        str: key path prefix
        WinRegistryFile: corresponding Windows Registry file or None if not
            available.
    """
    longest_key_path_prefix_upper = ''
    longest_key_path_prefix_length = len(longest_key_path_prefix_upper)
    for key_path_prefix_upper in self._registry_files:
      if key_path_upper.startswith(key_path_prefix_upper):
        key_path_prefix_length = len(key_path_prefix_upper)
        if key_path_prefix_length > longest_key_path_prefix_length:
          longest_key_path_prefix_upper = key_path_prefix_upper
          longest_key_path_prefix_length = key_path_prefix_length

    if not longest_key_path_prefix_upper:
      return None, None

    registry_file = self._registry_files.get(
        longest_key_path_prefix_upper, None)
    return longest_key_path_prefix_upper, registry_file

  def _GetCachedUserFileByPath(self, profile_path_upper):
    """Retrieves a cached user Windows Registry file for a profile path.

    Args:
      profile_path_upper (str): user profile path, in upper case.

    Returns:
      WinRegistryFile: corresponding Windows Registry file or None if not
          available.
    """
    return self._user_registry_files.get(profile_path_upper, None)

  def _GetCandidateFileMappingsByPath(self, key_path_upper):
    """Retrieves candidate Windows Registry file mappings for a specific path.

    Args:
      key_path_upper (str): Windows Registry key path, in upper case with
          a resolved root key alias.

    Returns:
      list[WinRegistryFileMapping]: candidate Windows Registry file mappings.
    """
    if key_path_upper[-1] == '\\':
      key_path_upper = key_path_upper[:-1]

    candidate_mappings = []
    for mapping in self._REGISTRY_FILE_MAPPINGS:
      if key_path_upper.startswith(mapping.key_path_prefix.upper()):
        candidate_mappings.append(mapping)

    return candidate_mappings

  def _GetFileByPath(self, key_path_upper):
    """Retrieves a Windows Registry file for a specific path.

    Args:
      key_path_upper (str): Windows Registry key path, in upper case with
          a resolved root key alias.

    Returns:
      tuple[st, WinRegistryFile]: upper case key path prefix and corresponding
          Windows Registry file or None if not available.
    """
    key_path_prefix, cached_registry_file = self._GetCachedFileByPath(
        key_path_upper)

    # NTUSER.DAT and UsrClass.dat have overlapping keys hence for key paths that
    # start "HKEY_CURRENT_USER\\Software\\Classes" do not assume the Windows
    # Registry file for "HKEY_CURRENT_USER" contains the classes key as well.

    if cached_registry_file and (
        key_path_prefix != 'HKEY_CURRENT_USER' or
        not key_path_upper.startswith('HKEY_CURRENT_USER\\SOFTWARE\\CLASSES')):
      return key_path_prefix, cached_registry_file

    key_path_prefix = ''
    candidate_mappings = self._GetCandidateFileMappingsByPath(key_path_upper)
    matching_mappings = []

    for mapping in candidate_mappings:
      mapping_key_path_prefix = mapping.key_path_prefix.upper()
      if mapping_key_path_prefix in self._registry_files:
        # Skip the Windows Registry file if it is already cached.
        continue

      try:
        registry_file = self._OpenFile(mapping.windows_path)
      except IOError:
        registry_file = None

      if len(key_path_prefix) < len(mapping_key_path_prefix):
        key_path_prefix = mapping_key_path_prefix

      if not registry_file:
        continue

      match = True
      if mapping.unique_key_paths:
        # If all unique key paths are found consider the file to match.
        for unique_key_path in mapping.unique_key_paths:
          registry_key = registry_file.GetKeyByPath(unique_key_path)
          if not registry_key:
            match = False

      if match:
        matching_mappings.append((mapping, registry_file))

    if not matching_mappings:
      return key_path_prefix, cached_registry_file

    key_path_prefix = None
    registry_file = None

    while matching_mappings:
      mapping, matching_registry_file = matching_mappings.pop(0)

      if (not key_path_prefix or
          len(mapping.key_path_prefix) > len(key_path_prefix)):
        if registry_file:
          registry_file.Close()

        key_path_prefix = mapping.key_path_prefix
        registry_file = matching_registry_file

    key_path_prefix_upper = key_path_prefix.upper()
    self.MapFile(key_path_prefix_upper, registry_file)

    return key_path_prefix_upper, registry_file

  def _GetKeyByPathFromFile(self, key_path):
    """Retrieves the key for a specific path from file.

    Args:
      key_path (str): Windows Registry key path.

    Returns:
      WinRegistryKey: Windows Registry key or None if not available.

    Raises:
      RuntimeError: if the key path prefix does not match the key path.
    """
    key_path_upper = key_path.upper()

    key_path_prefix_upper, registry_file = self._GetFileByPath(key_path_upper)
    if not registry_file:
      return None

    if not key_path_upper.startswith(key_path_prefix_upper):
      raise RuntimeError('Key path prefix mismatch.')

    key_path_suffix = key_path[len(key_path_prefix_upper):]
    relative_key_path = key_path_suffix or definitions.KEY_PATH_SEPARATOR

    return registry_file.GetKeyByPath(relative_key_path)

  def _GetRootVirtualKey(self):
    """Retrieves the root key.

    Returns:
      VirtualWinRegistryKey: Windows Registry root key.

    Raises:
      RuntimeError: if there are multiple matching mappings and
          the correct mapping cannot be resolved.
    """
    if not self._root_key:
      self._root_key = virtual.VirtualWinRegistryKey('')

      local_machine_key = None
      for sub_key_name in self._ROOT_SUB_KEYS:
        sub_registry_key = self.GetKeyByPath(sub_key_name)
        if not sub_registry_key:
          sub_registry_key = virtual.VirtualWinRegistryKey(
              sub_key_name, registry=self, relative_key_path=sub_key_name)

          if sub_key_name == 'HKEY_LOCAL_MACHINE':
            local_machine_key = sub_registry_key

        self._root_key.AddSubkey(sub_key_name, sub_registry_key)

      if local_machine_key:
        for sub_key_name in self._LOCAL_MACHINE_SUB_KEYS:
          sub_key_path = definitions.KEY_PATH_SEPARATOR.join([
              'HKEY_LOCAL_MACHINE', sub_key_name])
          sub_registry_key = self.GetKeyByPath(sub_key_path)
          if not sub_registry_key:
            sub_registry_key = virtual.VirtualWinRegistryKey(
                sub_key_name, registry=self, relative_key_path=sub_key_name)

          local_machine_key.AddSubkey(sub_key_name, sub_registry_key)

    return self._root_key

  def _GetUserFileByPath(self, path):
    """Retrieves an user Windows Registry file for a specific path.

    Args:
      path (str): path of the user Windows Registry file.

    Returns:
      WinRegistryFile: corresponding Windows Registry file or None if not
          available.
    """
    path_upper = path.upper()
    registry_file = self._GetCachedUserFileByPath(path_upper)
    if not registry_file:
      try:
        registry_file = self._OpenFile(path)
      except IOError:
        registry_file = None

    return registry_file

  def _GetUsersVirtualKey(self, key_path_suffix):
    """Virtual key callback to determine the users sub keys.

    Args:
      key_path_suffix (str): users Windows Registry key path suffix with
          leading path separator.

    Returns:
      WinRegistryKey: the users Windows Registry key or None if not available.

    Raises:
      RuntimeError: if the key path suffix is not supported.
    """
    if key_path_suffix[0] != definitions.KEY_PATH_SEPARATOR:
      raise RuntimeError(f'Unsupported key path suffix: {key_path_suffix:s}')

    user_key_name, _, key_path_suffix = key_path_suffix[1:].partition(
        definitions.KEY_PATH_SEPARATOR)
    user_key_name = user_key_name.upper()

    # HKEY_USERS\.DEFAULT is an alias for HKEY_USERS\S-1-5-18 which is
    # the Local System account.
    if user_key_name == '.DEFAULT':
      search_key_name = 'S-1-5-18'
    else:
      search_key_name = user_key_name

    if search_key_name.endswith('_CLASSES'):
      search_key_name = search_key_name[:-8]

    user_profile_list_key = self.GetKeyByPath(self._USER_PROFILE_LIST_KEY_PATH)
    if not user_profile_list_key:
      return None

    for user_profile_key in user_profile_list_key.GetSubkeys():
      if search_key_name == user_profile_key.name:
        profile_path_value = user_profile_key.GetValueByName('ProfileImagePath')
        if not profile_path_value:
          break

        profile_path = profile_path_value.GetDataAsObject()
        if not profile_path:
          break

        # HKEY_USERS\%SID%_CLASSES maps to UsrClass.dat
        if user_key_name.endswith('_CLASSES'):
          profile_path = definitions.KEY_PATH_SEPARATOR.join([
              profile_path, 'AppData', 'Local', 'Microsoft', 'Windows',
              'UsrClass.dat'])
        else:
          profile_path = definitions.KEY_PATH_SEPARATOR.join([
              profile_path, 'NTUSER.DAT'])

        registry_file = self._GetUserFileByPath(profile_path)
        if not registry_file:
          break

        key_path_prefix = definitions.KEY_PATH_SEPARATOR.join([
            'HKEY_USERS', user_key_name])
        key_path = definitions.KEY_PATH_SEPARATOR.join([
            key_path_prefix, key_path_suffix])

        registry_file.SetKeyPathPrefix(key_path_prefix)
        return registry_file.GetKeyByPath(key_path)

    return None

  def _GetVirtualKeyByPath(self, key_path):
    """Retrieves the virtual key for a specific path.

    Args:
      key_path (str): Windows Registry key path.

    Returns:
      VirtualWinRegistryKey: virtual Windows Registry key or None if not
          available.
    """
    key_path_upper = key_path.upper()

    for virtual_key_path, virtual_key_callback in self._VIRTUAL_KEYS:
      virtual_key_path_upper = virtual_key_path.upper()
      if key_path_upper.startswith(virtual_key_path_upper):
        key_path_suffix = key_path[len(virtual_key_path):]

        callback_function = getattr(self, virtual_key_callback)
        virtual_key = callback_function(key_path_suffix)
        if virtual_key:
          return virtual_key

    return None

  def _OpenFile(self, path):
    """Opens a Windows Registry file.

    Args:
      path (str): path of the Windows Registry file.

    Returns:
      WinRegistryFile: Windows Registry file or None if not available.
    """
    if not self._registry_file_reader:
      return None

    return self._registry_file_reader.Open(
        path, ascii_codepage=self._ascii_codepage)

  def GetKeyByPath(self, key_path):
    """Retrieves the key for a specific path.

    Args:
      key_path (str): Windows Registry key path.

    Returns:
      WinRegistryKey: Windows Registry key or None if not available.

    Raises:
      RuntimeError: if the root key is not supported or the key path prefix
          does not match the key path.
    """
    root_key_path, _, key_path = key_path.partition(
        definitions.KEY_PATH_SEPARATOR)

    # Resolve a root key alias.
    root_key_path = root_key_path.upper()
    root_key_path = self._ROOT_KEY_ALIASES.get(root_key_path, root_key_path)

    if root_key_path not in self._ROOT_SUB_KEYS:
      raise RuntimeError(f'Unsupported root subkey: {root_key_path:s}')

    key_path = definitions.KEY_PATH_SEPARATOR.join([root_key_path, key_path])

    registry_key = self._GetKeyByPathFromFile(key_path)
    if not registry_key:
      registry_key = self._GetVirtualKeyByPath(key_path)

    return registry_key

  def GetRegistryFileMapping(self, registry_file):
    """Determines the Registry file mapping based on the content of the file.

    Args:
      registry_file (WinRegistyFile): Windows Registry file.

    Returns:
      str: key path prefix or an empty string.

    Raises:
      RuntimeError: if there are multiple matching mappings and
          the correct mapping cannot be resolved.
    """
    if not registry_file:
      return ''

    candidate_mappings = []
    for mapping in self._REGISTRY_FILE_MAPPINGS:
      if not mapping.unique_key_paths:
        continue

      # If all unique key paths are found consider the file to match.
      match = True
      for key_path in mapping.unique_key_paths:
        registry_key = registry_file.GetKeyByPath(key_path)
        if not registry_key:
          match = False

      if match:
        candidate_mappings.append(mapping)

    if not candidate_mappings:
      return ''

    if len(candidate_mappings) == 1:
      return candidate_mappings[0].key_path_prefix

    key_path_prefixes = frozenset([
        mapping.key_path_prefix for mapping in candidate_mappings])

    expected_key_path_prefixes = frozenset([
        'HKEY_CURRENT_USER',
        'HKEY_CURRENT_USER\\Software\\Classes'])

    if key_path_prefixes == expected_key_path_prefixes:
      return 'HKEY_CURRENT_USER'

    raise RuntimeError('Unable to resolve Windows Registry file mapping.')

  def GetRootKey(self):
    """Retrieves the Windows Registry root key.

    Returns:
      VirtualWinRegistryKey: Windows Registry root key.
    """
    return self._GetRootVirtualKey()

  def MapFile(self, key_path_prefix, registry_file):
    """Maps the Windows Registry file to a specific key path prefix.

    Args:
      key_path_prefix (str): key path prefix.
      registry_file (WinRegistryFile): Windows Registry file.
    """
    key_path_prefix_upper = key_path_prefix.upper()

    self._registry_files[key_path_prefix_upper] = registry_file
    registry_file.SetKeyPathPrefix(key_path_prefix)

    # If HKEY_CURRENT_USER set HKEY_CURRENT_USER\\Software\\Classes as a virtual
    # key in the file.
    if key_path_prefix_upper == 'HKEY_CURRENT_USER' and isinstance(
        registry_file, regf.REGFWinRegistryFile):
      key_path_prefix_upper, usrclass_registry_file = self._GetFileByPath(
          self._USER_SOFTWARE_CLASSES_KEY_PATH)
      if (key_path_prefix_upper == self._USER_SOFTWARE_CLASSES_KEY_PATH and
          usrclass_registry_file is not None):
        registry_key = usrclass_registry_file.GetKeyByPath(
            '\\Software\\Classes')
        if registry_key and isinstance(registry_key, regf.REGFWinRegistryKey):
          # pylint: disable=protected-access
          pyregf_key = registry_key._pyregf_key
          registry_file.AddVirtualKey('\\Software\\Classes', pyregf_key)

    # If HKEY_LOCAL_MACHINE\\System set HKEY_LOCAL_MACHINE\\System\\
    # CurrentControlSet as a virtual key in the file.
    elif key_path_prefix_upper == 'HKEY_LOCAL_MACHINE\\SYSTEM' and isinstance(
        registry_file, regf.REGFWinRegistryFile):
      registry_file.AddCurrentControlSetKey()

  def MapUserFile(self, profile_path, registry_file):
    """Maps the user Windows Registry file to a specific profile path.

    Args:
      profile_path (str): profile path.
      registry_file (WinRegistryFile): user Windows Registry file.
    """
    self._user_registry_files[profile_path.upper()] = registry_file

  def OpenAndMapFile(self, path):
    """Opens Windows Registry file and maps it to its key path prefix.

    Args:
      path (str): path of the Windows Registry file.
    """
    registry_file = self._OpenFile(path)
    key_path_prefix = self.GetRegistryFileMapping(registry_file)
    self.MapFile(key_path_prefix, registry_file)

  def SplitKeyPath(self, key_path):
    """Splits the key path into path segments.

    Args:
      key_path (str): key path.

    Returns:
      list[str]: key path segments without the root path segment, which is an
          empty string.
    """
    return key_paths.SplitKeyPath(key_path)



# -*- coding: utf-8 -*-
"""Digital Forensics Windows Registry (dfWinReg).

dfWinReg, or Digital Forensics Windows Registry, is a Python module that
provides read-only access to Windows Registry objects.
"""

__version__ = '20240316'



# -*- coding: utf-8 -*-
"""This file contains the error classes."""


class Error(Exception):
  """Base error class."""


class WinRegistryValueError(Error):
  """Raised when a Windows Registry value cannot be read."""



# -*- coding: utf-8 -*-
"""A searcher to find keys and values within a Windows Registry."""

import re

try:
  import re._constants as sre_constants
except ImportError:
  import sre_constants  # pylint: disable=deprecated-module

from dfwinreg import decorators
from dfwinreg import glob2regex
from dfwinreg import key_paths


class FindSpec(object):
  """Find specification."""

  def __init__(
      self, key_path=None, key_path_glob=None, key_path_regex=None):
    """Initializes a find specification.

    Args:
      key_path (Optional[str|list[str]]): key path or key path segments,
          where None indicates no preference. The key path should be defined
          relative to the root of the Windows Registry. Note that the string
          will be split into segments based on the key path segment separator.
      key_path_glob (Optional[str:list[str]]): key path glob or key path glob
          segments, where None indicates no preference. The key path glob
          should be defined relative to the root of the Windows Registry. The
          default is None. Note that the string will be split into segments
          based on the key path segment separator.
      key_path_regex (Optional[str|list[str]]): key path regular expression or
          key path regular expression segments, where None indicates no
          preference. The key path regular expression should be defined
          relative to the root of the Windows Registry. The default is None.
          Note that the string will be split into segments based on the key
          path segment separator.

    Raises:
      TypeError: if the key_path, key_path_glob or key_path_regex type
          is not supported.
      ValueError: if the key_path, key_path_glob or key_path_regex arguments
          are used at the same time.
    """
    key_path_arguments = [argument for argument in (
        key_path, key_path_glob, key_path_regex) if argument]

    if len(key_path_arguments) > 1:
      raise ValueError((
          'The key_path, key_path_glob and key_path_regex arguments cannot '
          'be used at same time.'))

    super(FindSpec, self).__init__()
    self._is_regex = False
    self._key_path_segments = None
    self._number_of_key_path_segments = 0

    if key_path is not None:
      if isinstance(key_path, str):
        self._key_path_segments = key_paths.SplitKeyPath(key_path)
      elif isinstance(key_path, list):
        self._key_path_segments = key_path
      else:
        key_path_type = type(key_path)
        raise TypeError(f'Unsupported key path type: {key_path_type!s}.')

    elif key_path_glob is not None:
      # The regular expression from glob2regex contains escaped forward
      # slashes "/", which needs to be undone.

      if isinstance(key_path_glob, str):
        key_path_regex = glob2regex.Glob2Regex(key_path_glob)
        key_path_regex = key_path_regex.replace('\\/', '/')

        # The backslash '\' is escaped within a regular expression.
        self._key_path_segments = key_paths.SplitKeyPath(
            key_path_regex, path_separator='\\\\')

      elif isinstance(key_path_glob, list):
        self._key_path_segments = []
        for key_path_segment in key_path_glob:
          key_path_regex = glob2regex.Glob2Regex(key_path_segment)
          key_path_regex = key_path_regex.replace('\\/', '/')

          self._key_path_segments.append(key_path_regex)

      else:
        key_path_glob_type = type(key_path_glob)
        raise TypeError(
            f'Unsupported key_path_glob type: {key_path_glob_type!s}.')

      self._is_regex = True

    elif key_path_regex is not None:
      if isinstance(key_path_regex, str):
        # The backslash '\' is escaped within a regular expression.
        self._key_path_segments = key_paths.SplitKeyPath(
            key_path_regex, path_separator='\\\\')
      elif isinstance(key_path_regex, list):
        self._key_path_segments = key_path_regex
      else:
        key_path_regex_type = type(key_path_regex)
        raise TypeError(
            f'Unsupported key_path_regex type: {key_path_regex_type!s}.')

      self._is_regex = True

    if self._key_path_segments is not None:
      self._number_of_key_path_segments = len(self._key_path_segments)

  def _CheckKeyPath(self, registry_key, search_depth):
    """Checks the key path find specification.

    Args:
      registry_key (WinRegistryKey): Windows Registry key.
      search_depth (int): number of key path segments to compare.

    Returns:
      bool: True if the Windows Registry key matches the find specification,
          False if not.
    """
    if self._key_path_segments is None:
      return False

    if search_depth < 0 or search_depth > self._number_of_key_path_segments:
      return False

    # Note that the root has no entry in the key path segments and
    # no name to match.
    if search_depth == 0:
      segment_name = ''
    else:
      segment_name = self._key_path_segments[search_depth - 1]

      if self._is_regex:
        if isinstance(segment_name, str):
          # Allow '\n' to be matched by '.' and make '\w', '\W', '\b', '\B',
          # '\d', '\D', '\s' and '\S' Unicode safe.
          flags = re.DOTALL | re.IGNORECASE | re.UNICODE

          try:
            segment_name = re.compile(f'^{segment_name:s}$', flags=flags)
          except sre_constants.error:
            # TODO: set self._key_path_segments[search_depth - 1] to None ?
            return False

          self._key_path_segments[search_depth - 1] = segment_name

      else:
        segment_name = segment_name.upper()
        self._key_path_segments[search_depth - 1] = segment_name

    if search_depth > 0:
      key_name = registry_key.path.rsplit('\\', maxsplit=1)[-1]
      if self._is_regex:
        # pylint: disable=no-member
        if not segment_name.match(key_name):
          return False

      elif segment_name != key_name.upper():
        return False

    return True

  def _CompareWithKeyPathSegment(self, key_path_segment, segment_index):
    """Compares a key path segment against a find specification.

    Args:
      key_path_segment (str): key path segment.
      segment_index (int): index of the key path segment to compare against,
          where 0 represents the root segment.

    Returns:
      bool: True if the key path segment of the Windows Registry key matches
          that of the find specification, False if not or if the find
          specification has no key path defined.
    """
    if (self._key_path_segments is None or segment_index < 0 or
        segment_index > self._number_of_key_path_segments):
      return False

    segment_name = self._key_path_segments[segment_index]

    if self._is_regex:
      if isinstance(segment_name, str):
        # Allow '\n' to be matched by '.' and make '\w', '\W', '\b', '\B',
        # '\d', '\D', '\s' and '\S' Unicode safe.
        flags = re.DOTALL | re.UNICODE | re.IGNORECASE

        try:
          segment_name = re.compile(f'^{segment_name:s}$', flags=flags)
        except sre_constants.error:
          # TODO: set self._key_path_segments[segment_index] to None ?
          return False

        self._key_path_segments[segment_index] = segment_name

    else:
      segment_name = segment_name.upper()
      self._key_path_segments[segment_index] = segment_name

    if self._is_regex:
      result = bool(segment_name.match(key_path_segment))  # pylint: disable=no-member
      return result

    result = bool(segment_name == key_path_segment.upper())
    return result

  def AtLastKeyPathSegment(self, segment_index):
    """Determines if the a key path segment is the last one or greater.

    Args:
      segment_index (int): index of the key path segment.

    Returns:
      bool: True if at maximum depth, False if not.
    """
    return bool(self._key_path_segments is not None and
                (segment_index + 1) >= self._number_of_key_path_segments)

  @decorators.deprecated
  def AtMaximumDepth(self, search_depth):
    """Determines if the find specification is at maximum depth.

    This method is deprecated use AtLastKeyPathSegment instead.

    Args:
      search_depth (int): number of key path segments to compare.

    Returns:
      bool: True if at maximum depth, False if not.
    """
    if self._key_path_segments is not None:
      if search_depth >= self._number_of_key_path_segments:
        return True

    return False

  def CompareKeyPath(self, registry_key):
    """Compares a Windows Registry key path against the find specification.

    Args:
      registry_key (WinRegistryKey): Windows Registry key.

    Returns:
      bool: True if the key path of the Windows Registry key matches that of
          the find specification, False if not or if the find specification
          has no key path defined.
    """
    key_path = getattr(registry_key, 'path', None)
    if self._key_path_segments is None or key_path is None:
      return False

    key_path_segments = key_path.split('\\')

    for segment_index in range(self._number_of_key_path_segments):
      try:
        key_path_segment = key_path_segments[segment_index]
      except IndexError:
        return False

      if not self._CompareWithKeyPathSegment(key_path_segment, segment_index):
        return False

    return True

  def CompareNameWithKeyPathSegment(self, registry_key, segment_index):
    """Compares a Windows Registry key name against a key path segment.

    Args:
      registry_key (WinRegistryKey): Windows Registry key.
      segment_index (int): index of the key path segment to compare against,
          where 0 represents the root segment.

    Returns:
      bool: True if the key path segment of the Windows Registry key matches
          that of the find specification, False if not or if the find
          specification has no key path defined.
    """
    key_name = registry_key.path.rsplit('\\', maxsplit=1)[-1]
    return self._CompareWithKeyPathSegment(key_name, segment_index)

  def HasKeyPath(self):
    """Determines if the find specification has a key path defined.

    Returns:
      bool: True if find specification has a key path defined, False if not.
    """
    return bool(self._key_path_segments)

  def IsLastKeyPathSegment(self, segment_index):
    """Determines if the a key path segment is the last one.

    Args:
      segment_index (int): index of the key path path segment.

    Returns:
      bool: True if at maximum depth, False if not.
    """
    return bool(self._key_path_segments is not None and
                (segment_index + 1) == self._number_of_key_path_segments)

  @decorators.deprecated
  def Matches(self, registry_key, search_depth):
    """Determines if the Windows Registry key matches the find specification.

    This method is deprecated use CompareKeyPath or
    CompareNameWithKeyPathSegment instead.

    Args:
      registry_key (WinRegistryKey): Windows Registry key.
      search_depth (int): number of key path segments to compare.

    Returns:
      tuple: containing:

        bool: True if the Windows Registry key matches the find specification,
            False otherwise.
        bool: True if the key path matches, False if not or None if no key path
            specified.
    """
    if self._key_path_segments is None:
      key_path_match = None
    else:
      key_path_match = self._CheckKeyPath(registry_key, search_depth)
      if not key_path_match:
        return False, key_path_match

      if search_depth != self._number_of_key_path_segments:
        return False, key_path_match

    return True, key_path_match


class WinRegistrySearcher(object):
  """Searcher for key and values within a Windows Registry."""

  def __init__(self, win_registry):
    """Initializes a Windows Registry searcher.

    Args:
      win_registry (WinRegistry): Windows Registry.

    Raises:
      ValueError: when Windows Registry is not set.
    """
    if not win_registry:
      raise ValueError('Missing Windows Registry value.')

    super(WinRegistrySearcher, self).__init__()
    self._win_registry = win_registry

  def _FindInKey(self, registry_key, find_specs, segment_index):
    """Searches for matching keys within the Windows Registry key.

    Args:
      registry_key (WinRegistryKey): Windows Registry key.
      find_specs (list[FindSpec]): find specifications.
      segment_index (int): index of the key path segment to compare.

    Yields:
      str: key path of a matching Windows Registry key.
    """
    sub_find_specs = []
    for find_spec in find_specs:
      has_key_path = find_spec.HasKeyPath()
      # Do a quick check to see if the current key path segment matches.
      key_path_match = find_spec.CompareNameWithKeyPathSegment(
          registry_key, segment_index)
      is_last_key_path_segment = find_spec.IsLastKeyPathSegment(segment_index)

      if key_path_match and is_last_key_path_segment:
        # Check if the full key path matches.
        key_path_match = find_spec.CompareKeyPath(registry_key)

      if not has_key_path or (key_path_match and is_last_key_path_segment):
        # TODO: add support for CompareTraits.
        yield registry_key.path

      at_last_key_path_segment = find_spec.AtLastKeyPathSegment(segment_index)
      if (not has_key_path or key_path_match) and not at_last_key_path_segment:
        sub_find_specs.append(find_spec)

    if sub_find_specs:
      segment_index += 1
      for sub_registry_key in registry_key.GetSubkeys():
        for matching_key_path in self._FindInKey(
            sub_registry_key, sub_find_specs, segment_index):
          yield matching_key_path

  def Find(self, find_specs=None):
    """Searches for matching keys within the Windows Registry.

    Args:
      find_specs (list[FindSpec]): find specifications. where None
          will return all allocated Windows Registry keys.

    Yields:
      str: key path of a matching Windows Registry key.
    """
    if not find_specs:
      find_specs = [FindSpec()]

    root_registry_key = self._win_registry.GetRootKey()
    for sub_registry_key in root_registry_key.GetSubkeys():
      for matching_path in self._FindInKey(sub_registry_key, find_specs, 0):
        yield matching_path

  def GetKeyByPath(self, key_path):
    """Retrieves a Windows Registry key for a path specification.

    Args:
      key_path (str): key path.

    Returns:
      WinRegistryKey: Windows Registry key or None.
    """
    return self._win_registry.GetKeyByPath(key_path)

  def SplitKeyPath(self, key_path):
    """Splits the key path into path segments.

    Args:
      key_path (str): key path.

    Returns:
      list[str]: key path segments without the root path segment, which is an
          empty string.
    """
    return self._win_registry.SplitKeyPath(key_path)



# -*- coding: utf-8 -*-
"""The Windows Registry object interfaces."""

import abc

from dfwinreg import definitions
from dfwinreg import key_paths


class WinRegistryFile(object):
  """Windows Registry file interface."""

  def __init__(self, ascii_codepage='cp1252', key_path_prefix=''):
    """Initializes a Windows Registry file.

    Args:
      ascii_codepage (Optional[str]): ASCII string codepage.
      key_path_prefix (Optional[str]): Windows Registry key path prefix.
    """
    super(WinRegistryFile, self).__init__()
    self._ascii_codepage = ascii_codepage
    self._key_path_prefix = key_path_prefix
    self._key_path_prefix_length = len(key_path_prefix)
    self._key_path_prefix_upper = key_path_prefix.upper()

  @abc.abstractmethod
  def Close(self):
    """Closes the Windows Registry file."""

  @abc.abstractmethod
  def GetKeyByPath(self, key_path):
    """Retrieves the key for a specific path.

    Args:
      key_path (str): Windows Registry key path.

    Returns:
      WinRegistryKey: Windows Registry key or None if not available.
    """

  @abc.abstractmethod
  def GetRootKey(self):
    """Retrieves the root key.

    Returns:
      WinRegistryKey: Windows Registry root key or None if not available.
    """

  @abc.abstractmethod
  def Open(self, file_object):
    """Opens the Windows Registry file using a file-like object.

    Args:
      file_object (file): file-like object.

    Returns:
      bool: True if successful or False if not.
    """

  def RecurseKeys(self):
    """Recurses the Windows Registry keys starting with the root key.

    Yields:
      WinRegistryKey: Windows Registry key.
    """
    root_key = self.GetRootKey()
    if root_key:
      for registry_key in root_key.RecurseKeys():
        yield registry_key

  def SetKeyPathPrefix(self, key_path_prefix):
    """Sets the Window Registry key path prefix.

    Args:
      key_path_prefix (str): Windows Registry key path prefix.
    """
    self._key_path_prefix = key_path_prefix
    self._key_path_prefix_length = len(key_path_prefix)
    self._key_path_prefix_upper = key_path_prefix.upper()


class WinRegistryFileReader(object):
  """Windows Registry file reader interface."""

  @abc.abstractmethod
  def Open(self, path, ascii_codepage='cp1252'):
    """Opens a Windows Registry file specified by the path.

    Args:
      path (str): path of the Windows Registry file. The path is a Windows
          path relative to the root of the file system that contains the
          specific Windows Registry file, for example:
          C:\\Windows\\System32\\config\\SYSTEM
      ascii_codepage (Optional[str]): ASCII string codepage.

    Returns:
      WinRegistryFile: Windows Registry file or None.
    """


class WinRegistryKey(object):
  """Windows Registry key interface."""

  def __init__(self, key_helper=None, key_path_prefix='', relative_key_path=''):
    """Initializes a Windows Registry key.

    Args:
      key_helper (Optional[WinRegistryKeyHelper]): Windows Registry key helper.
      key_path_prefix (Optional[str]): Windows Registry key path prefix.
      relative_key_path (Optional[str]): relative Windows Registry key path.
    """
    super(WinRegistryKey, self).__init__()
    self._key_helper = key_helper
    self._key_path_prefix = key_path_prefix
    self._relative_key_path = relative_key_path

  @property
  @abc.abstractmethod
  def class_name(self):
    """str: class name of the key or None if not available."""

  @property
  @abc.abstractmethod
  def last_written_time(self):
    """dfdatetime.DateTimeValues: last written time or None."""

  @property
  @abc.abstractmethod
  def name(self):
    """str: name of the key."""

  @property
  @abc.abstractmethod
  def number_of_subkeys(self):
    """int: number of subkeys within the key."""

  @property
  @abc.abstractmethod
  def number_of_values(self):
    """int: number of values within the key."""

  @property
  @abc.abstractmethod
  def offset(self):
    """int: offset of the key within the Windows Registry file or None."""

  @property
  def path(self):
    """str: Windows Registry key path."""
    return key_paths.JoinKeyPath([
        self._key_path_prefix, self._relative_key_path])

  @abc.abstractmethod
  def GetSubkeyByIndex(self, index):
    """Retrieves a subkey by index.

    Args:
      index (int): index of the subkey.

    Returns:
      WinRegistryKey: Windows Registry subkey.

    Raises:
      IndexError: if the index is out of bounds.
    """

  @abc.abstractmethod
  def GetSubkeyByName(self, name):
    """Retrieves a subkey by name.

    Args:
      name (str): name of the subkey.

    Returns:
      WinRegistryKey: Windows Registry subkey or None if not found.
    """

  @abc.abstractmethod
  def GetSubkeyByPath(self, key_path):
    """Retrieves a subkey by a path.

    Args:
      key_path (str): relative key path of the subkey.

    Returns:
      WinRegistryKey: Windows Registry subkey or None if not found.
    """

  @abc.abstractmethod
  def GetSubkeys(self):
    """Retrieves all subkeys within the key.

    Yields:
      WinRegistryKey: Windows Registry subkey.
    """

  @abc.abstractmethod
  def GetValueByName(self, name):
    """Retrieves a value by name.

    Args:
      name (str): name of the value or an empty string for the default value.

    Returns:
      WinRegistryValue: Windows Registry value or None if not found.
    """

  @abc.abstractmethod
  def GetValues(self):
    """Retrieves all values within the key.

    Yields:
      WinRegistryValue: Windows Registry value.
    """

  def RecurseKeys(self):
    """Recurses the subkeys starting with the key.

    Yields:
      WinRegistryKey: Windows Registry key.
    """
    yield self
    for subkey in self.GetSubkeys():
      for key in subkey.RecurseKeys():
        yield key


class WinRegistryKeyHelper(object):
  """Windows Registry key helper."""


class WinRegistryValue(object):
  """Windows Registry value interface."""

  _DATA_TYPE_STRINGS = {
      0: 'REG_NONE',
      1: 'REG_SZ',
      2: 'REG_EXPAND_SZ',
      3: 'REG_BINARY',
      4: 'REG_DWORD_LE',
      5: 'REG_DWORD_BE',
      6: 'REG_LINK',
      7: 'REG_MULTI_SZ',
      8: 'REG_RESOURCE_LIST',
      9: 'REG_FULL_RESOURCE_DESCRIPTOR',
      10: 'REG_RESOURCE_REQUIREMENTS_LIST',
      11: 'REG_QWORD'}

  _STRING_VALUE_TYPES = frozenset([
      definitions.REG_SZ, definitions.REG_EXPAND_SZ, definitions.REG_LINK])

  @property
  @abc.abstractmethod
  def data(self):
    """bytes: value data."""

  @property
  @abc.abstractmethod
  def data_type(self):
    """int: data type."""

  @property
  def data_type_string(self):
    """str: string representation of the data type."""
    return self._DATA_TYPE_STRINGS.get(self.data_type, 'UNKNOWN')

  @property
  @abc.abstractmethod
  def name(self):
    """str: name of the value."""

  @property
  @abc.abstractmethod
  def offset(self):
    """int: offset of the value within the Windows Registry file."""

  def DataIsBinaryData(self):
    """Determines, based on the data type, if the data is binary data.

    The data types considered binary data are: REG_BINARY.

    Returns:
      bool: True if the data is a binary data, False otherwise.
    """
    return self.data_type == definitions.REG_BINARY

  def DataIsInteger(self):
    """Determines, based on the data type, if the data is an integer.

    The data types considered strings are: REG_DWORD (REG_DWORD_LITTLE_ENDIAN),
    REG_DWORD_BIG_ENDIAN and REG_QWORD.

    Returns:
      bool: True if the data is an integer, False otherwise.
    """
    return self.data_type in (
        definitions.REG_DWORD, definitions.REG_DWORD_BIG_ENDIAN,
        definitions.REG_QWORD)

  def DataIsMultiString(self):
    """Determines, based on the data type, if the data is a multi string.

    The data types considered multi strings are: REG_MULTI_SZ.

    Returns:
      bool: True if the data is multi string, False otherwise.
    """
    return self.data_type == definitions.REG_MULTI_SZ

  def DataIsString(self):
    """Determines, based on the data type, if the data is a string.

    The data types considered strings are: REG_SZ and REG_EXPAND_SZ.

    Returns:
      bool: True if the data is a string, False otherwise.
    """
    return self.data_type in (definitions.REG_SZ, definitions.REG_EXPAND_SZ)

  @abc.abstractmethod
  def GetDataAsObject(self):
    """Retrieves the data as an object.

    Returns:
      object: data as a Python type.
    """



# -*- coding: utf-8 -*-
"""Windows 9x/Me Registry (CREG) objects implementation using pycreg."""

from dfdatetime import semantic_time as dfdatetime_semantic_time

import pycreg

from dfwinreg import definitions
from dfwinreg import errors
from dfwinreg import interface
from dfwinreg import key_paths


class CREGWinRegistryFile(interface.WinRegistryFile):
  """Implementation of a Windows Registry file using pycreg."""

  def __init__(self, ascii_codepage='cp1252', key_path_prefix=''):
    """Initializes the Windows Registry file.

    Args:
      ascii_codepage (Optional[str]): ASCII string codepage.
      key_path_prefix (Optional[str]): Windows Registry key path prefix.
    """
    super(CREGWinRegistryFile, self).__init__(
        ascii_codepage=ascii_codepage, key_path_prefix=key_path_prefix)
    self._file_object = None
    self._creg_file = pycreg.file()
    self._creg_file.set_ascii_codepage(ascii_codepage)

  def Close(self):
    """Closes the Windows Registry file."""
    self._creg_file.close()
    self._file_object = None

  def GetKeyByPath(self, key_path):
    """Retrieves the key for a specific path.

    Args:
      key_path (str): Windows Registry key path.

    Returns:
      WinRegistryKey: Windows Registry key or None if not available.
    """
    key_path_upper = key_path.upper()
    if key_path_upper.startswith(self._key_path_prefix_upper):
      relative_key_path = key_path[self._key_path_prefix_length:]
    elif key_path.startswith(definitions.KEY_PATH_SEPARATOR):
      relative_key_path = key_path
      key_path = ''.join([self._key_path_prefix, key_path])
    else:
      return None

    try:
      creg_key = self._creg_file.get_key_by_path(relative_key_path)
    except IOError:
      creg_key = None

    if not creg_key:
      return None

    return CREGWinRegistryKey(
        creg_key, key_path_prefix=self._key_path_prefix,
        relative_key_path=relative_key_path)

  def GetRootKey(self):
    """Retrieves the root key.

    Returns:
      WinRegistryKey: Windows Registry root key or None if not available.
    """
    creg_key = self._creg_file.get_root_key()
    if not creg_key:
      return None

    return CREGWinRegistryKey(
        creg_key, key_path_prefix=self._key_path_prefix,
        relative_key_path='')

  def Open(self, file_object):
    """Opens the Windows Registry file using a file-like object.

    Args:
      file_object (file): file-like object.

    Returns:
      bool: True if successful or False if not.
    """
    self._file_object = file_object
    self._creg_file.open_file_object(self._file_object)
    return True


class CREGWinRegistryKey(interface.WinRegistryKey):
  """Implementation of a Windows Registry key using pycreg."""

  def __init__(
      self, pycreg_key, key_helper=None, key_path_prefix='',
      relative_key_path=''):
    """Initializes a Windows Registry key.

    Args:
      pycreg_key (pycreg.key): pycreg key object.
      key_helper (Optional[WinRegistryKeyHelper]): Windows Registry key helper.
      key_path_prefix (Optional[str]): Windows Registry key path prefix.
      relative_key_path (Optional[str]): relative Windows Registry key path.
    """
    super(CREGWinRegistryKey, self).__init__(
        key_helper=key_helper, key_path_prefix=key_path_prefix,
        relative_key_path=relative_key_path)
    self._pycreg_key = pycreg_key

  @property
  def class_name(self):
    """str: class name of the key or None if not available."""
    return None

  @property
  def last_written_time(self):
    """dfdatetime.DateTimeValues: last written time."""
    return dfdatetime_semantic_time.SemanticTime('Not set')

  @property
  def name(self):
    """str: name of the key."""
    return self._pycreg_key.name

  @property
  def number_of_subkeys(self):
    """int: number of subkeys within the key."""
    return self._pycreg_key.number_of_sub_keys

  @property
  def number_of_values(self):
    """int: number of values within the key."""
    return self._pycreg_key.number_of_values

  @property
  def offset(self):
    """int: offset of the key within the Windows Registry file or None."""
    return self._pycreg_key.offset

  def GetSubkeyByIndex(self, index):
    """Retrieves a subkey by index.

    Args:
      index (int): index of the subkey.

    Returns:
      WinRegistryKey: Windows Registry subkey.

    Raises:
      IndexError: if the index is out of bounds.
    """
    if index < 0 or index >= self._pycreg_key.number_of_sub_keys:
      raise IndexError('Index out of bounds.')

    pycreg_key = self._pycreg_key.get_sub_key(index)
    relative_key_path = key_paths.JoinKeyPath([
        self._relative_key_path, pycreg_key.name])
    return CREGWinRegistryKey(
        pycreg_key, key_helper=self._key_helper,
        key_path_prefix=self._key_path_prefix,
        relative_key_path=relative_key_path)

  def GetSubkeyByName(self, name):
    """Retrieves a subkey by name.

    Args:
      name (str): name of the subkey.

    Returns:
      WinRegistryKey: Windows Registry subkey or None if not found.
    """
    pycreg_key = self._pycreg_key.get_sub_key_by_name(name)
    if not pycreg_key:
      return None

    relative_key_path = key_paths.JoinKeyPath([
        self._relative_key_path, pycreg_key.name])
    return CREGWinRegistryKey(
        pycreg_key, key_helper=self._key_helper,
        key_path_prefix=self._key_path_prefix,
        relative_key_path=relative_key_path)

  def GetSubkeyByPath(self, key_path):
    """Retrieves a subkey by path.

    Args:
      key_path (str): path of the subkey.

    Returns:
      WinRegistryKey: Windows Registry subkey or None if not found.
    """
    pycreg_key = self._pycreg_key.get_sub_key_by_path(key_path)
    if not pycreg_key:
      return None

    relative_key_path = key_paths.JoinKeyPath([
        self._relative_key_path, key_path])
    return CREGWinRegistryKey(
        pycreg_key, key_helper=self._key_helper,
        key_path_prefix=self._key_path_prefix,
        relative_key_path=relative_key_path)

  def GetSubkeys(self):
    """Retrieves all subkeys within the key.

    Yields:
      WinRegistryKey: Windows Registry subkey.
    """
    for pycreg_key in self._pycreg_key.sub_keys:
      relative_key_path = key_paths.JoinKeyPath([
          self._relative_key_path, pycreg_key.name])
      yield CREGWinRegistryKey(
          pycreg_key, key_helper=self._key_helper,
          key_path_prefix=self._key_path_prefix,
          relative_key_path=relative_key_path)

  def GetValueByName(self, name):
    """Retrieves a value by name.

    Value names are not unique and pycreg provides first match for the value.

    Args:
      name (str): name of the value or an empty string for the default value.

    Returns:
      WinRegistryValue: Windows Registry value if a corresponding value was
          found or None if not.
    """
    pycreg_value = self._pycreg_key.get_value_by_name(name)
    if not pycreg_value:
      return None

    return CREGWinRegistryValue(pycreg_value)

  def GetValues(self):
    """Retrieves all values within the key.

    Yields:
      WinRegistryValue: Windows Registry value.
    """
    for pycreg_value in self._pycreg_key.values:
      yield CREGWinRegistryValue(pycreg_value)


class CREGWinRegistryValue(interface.WinRegistryValue):
  """Implementation of a Windows Registry value using pycreg."""

  # Note that missing-return-doc is broken for pylint 1.7.x
  # pylint: disable=missing-return-doc

  def __init__(self, pycreg_value):
    """Initializes a Windows Registry value.

    Args:
      pycreg_value (pycreg.value): pycreg value object.
    """
    super(CREGWinRegistryValue, self).__init__()
    self._pycreg_value = pycreg_value

  # Pylint 1.7.x seems to be get confused about properties.
  # pylint: disable=missing-return-type-doc
  @property
  def data(self):
    """bytes: value data as a byte string.

    Raises:
      WinRegistryValueError: if the value data cannot be read.
    """
    try:
      return self._pycreg_value.data
    except IOError as exception:
      raise errors.WinRegistryValueError((
          f'Unable to read data from value: {self._pycreg_value.name:s} '
          f'with error: {exception!s}'))

  @property
  def data_type(self):
    """int: data type."""
    return self._pycreg_value.type

  @property
  def name(self):
    """str: name of the value."""
    return self._pycreg_value.name

  @property
  def offset(self):
    """int: offset of the value within the Windows Registry file."""
    return self._pycreg_value.offset

  def GetDataAsObject(self):
    """Retrieves the data as an object.

    Returns:
      object: data as a Python type.

    Raises:
      WinRegistryValueError: if the value data cannot be read.
    """
    try:
      if self._pycreg_value.type in self._STRING_VALUE_TYPES:
        value_data = self._pycreg_value.get_data_as_string()

      elif self._pycreg_value.type in definitions.INTEGER_VALUE_TYPES:
        value_data = self._pycreg_value.get_data_as_integer()

      elif self._pycreg_value.type == definitions.REG_MULTI_SZ:
        value_data = self._pycreg_value.get_data_as_multi_string()

      else:
        value_data = self._pycreg_value.data

    except (IOError, OverflowError) as exception:
      raise errors.WinRegistryValueError((
          f'Unable to read data from value: {self._pycreg_value.name:s} '
          f'with error: {exception!s}'))

    return value_data
